  // AC
  memset((uint8_t*) &adc.ac, 0, sizeof(adc.ac));

  // remove offsets
  struct powerReadProto *p_buffer = &adc.buffer[0];
  while (p_buffer < &adc.buffer[ADC_POWERREAD_SAMPLES]) {
    p_buffer->voltage -= adc.calibration.ac.voltage;
    p_buffer->current -= adc.calibration.ac.current;
    p_buffer++;
  }

  // find zero cross points
  struct powerReadProto* zerocross[4];
  uint8_t isHigh = false;

  // voltage zero cross
  p_buffer = &adc.buffer[0];
  for (uint8_t i=0; i<2; i++) {
    // find next high point
    while (p_buffer < &adc.buffer[ADC_POWERREAD_SAMPLES]) {
      if (p_buffer->voltage > ADC_ZEROCROSS_OFFSET_HIGH) {
        isHigh = true;
        break;
      }
      p_buffer++;
    }
    if (! isHigh) {
      adcSampleDc();
      return;
    }
    while (p_buffer < &adc.buffer[ADC_POWERREAD_SAMPLES]) {
      p_buffer++;
      if (p_buffer->voltage <= 0) {
        isHigh = false;
        zerocross[i] = p_buffer;
        break;
      }
    }
    if (isHigh) {
      adcSampleDc();
      return;
    }
  }

  // voltage samples
  uint16_t samples = zerocross[1] - zerocross[0];

  // frequency
  adc.ac.frequency.value = (ADC_FREQUENCY*10U) / samples; // us to hz

  // accumulate
  struct {
    uint64_t voltage;
    uint64_t current;
  } accumulators = { 0, 0 };
  p_buffer = zerocross[0];
  while (p_buffer <= zerocross[1]) {
    accumulators.voltage += (int32_t) p_buffer->voltage * p_buffer->voltage;
    accumulators.current += (int32_t) p_buffer->current * p_buffer->current;
    p_buffer++;
  }
  
  // apparent voltage
  uint32_t value = accumulators.voltage / samples;
  value = sqrt32(value);
  value *= (ADC_CVT_MULT*10U*ADC_AC_VOLTAGE_RESISTOR_A);
  value /= ADC_CVT_DIV;
  adc.ac.power.apparent.voltage = value;

  // apparent current
  value = accumulators.current / samples;
  value = sqrt32(value);
  value *= ADC_CVT_MULT*ADC_AC_CURRENT_MULTI;
  value /=  ADC_CVT_DIV*ADC_AC_CURRENT_SENSITIVITY;
  value /= ADC_CVT_DIV;
  adc.ac.power.apparent.current = value;
  adc.ac.power.apparent.wattage = (value * adc.ac.power.apparent.voltage) / 1000U;

  // current zero cross
  p_buffer = (struct powerReadProto*) zerocross[0] + 1;
  for (uint8_t i=2; i<4; i++) {
    // find high point
    while (p_buffer < &adc.buffer[ADC_POWERREAD_SAMPLES]) {
      if (p_buffer->current > ADC_ZEROCROSS_OFFSET_HIGH) {
        isHigh = true;
        break;
      }
      p_buffer++;
    }
    if (! isHigh) {
      adcSampleDc();
      return;
    }
    while (p_buffer < &adc.buffer[ADC_POWERREAD_SAMPLES]) {
      p_buffer++;
      if (p_buffer->current <= 0) {
        isHigh = false;
        zerocross[i] = p_buffer;
        break;
      }
    }
    if (isHigh) {
      adcSampleDc();
      return;
    }
  }

  // accumulate current
  accumulators.current = 0;
  p_buffer = zerocross[2];
  while (p_buffer <= zerocross[3]) {
    accumulators.current += (int32_t) p_buffer->current * p_buffer->current;
    p_buffer++;
  }

  // real power
  samples = zerocross[3] - zerocross[2];
  value = accumulators.current / samples;
  value = sqrt32(value);
  value *= ADC_CVT_MULT*ADC_AC_CURRENT_MULTI;
  value /=  ADC_CVT_DIV*ADC_AC_CURRENT_SENSITIVITY;
  adc.ac.power.real.wattage = (value * adc.ac.power.apparent.voltage) / 1000U;

  value = (uint32_t) adc.ac.power.apparent.wattage * 100U;
  adc.ac.power.factor = !adc.ac.power.real.wattage? 0 : value / adc.ac.power.real.wattage;
